#!/usr/bin/perl

# Author : Brian J. Murrell (brian@interlinx.bc.ca)
# Based heavily on Shell.pm (in fact I ripped it off completely!)
#
# Contributors : 
#
# Copyright (C) 2003 Brian J. Murrell (brian@interlinx.bc.ca)
# Copyright (C) 2004 Damien KROTKINE (damien@tuxfamily.org)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


=head1 NAME

PERL : Libconf::Glueconf::Generic::KeyValue
C/PYTHON/RUBY : Generic/KeyValue

=head1 DESCRIPTION

Libconf::Glueconf::Generic::KeyValue is a generic template that handles config files that contain semantic informations of type : (KEY - VALUE).

=head1 SYNOPSIS

  $autologin = new Libconf::Glueconf::Generic::KeyValue({ filename => 'autologin',
                                                          separator_char => '=',
                                                          allow_space => 0,
                                                          comments_struct => [['#']],
                                                          handle_quote => 1,
                                                          handle_multiples_lines => 0,
                                                        });
  $autologin->{USER} = 'foo';
  $autologin->write_conf('output_file');

=head1 CONSTRUCTOR

=over

=item B<new($options)>

creates the template

  $template = new Libconf::Templates::Generic::KeyValue({
                                                         .. arguments .. 
                                                        })

the options are the same as in L<Libconf::Templates::Generic::KeyValue>, look
there to have the complete list.

=back

=head1 METHODS

See <Libconf::Glueconf> for general methods

=cut

package Libconf::Glueconf::Generic::KeyValue;
use strict;
use Libconf;
use Libconf::Glueconf;
use Libconf::Templates::Generic::KeyValue;
our @ISA = qw(Libconf::Glueconf);

sub new {
    my ($class, $args, $data_description, $data_mapping) = @_;
    my $libconf = new Libconf::Templates::Generic::KeyValue(init_args($args, { glueconf_template_name => 'Generic/KeyValue' } ));
    $libconf->set_uniq(qw(key));
    $libconf->read_conf();
    tie my %wrapper, 'Libconf::Glueconf::Generic::KeyValue::Wrapper', $libconf, $data_description, $data_mapping;
    bless \%wrapper, $class;
}

sub freeze {
    my ($self) = @_;
    tied(%$self)->{_frozen} = 1;
}

sub thaw {
    my ($self) = @_;
    my $tied = tied(%$self);
    # we delete from end to start to respect indexes.
    $tied->{libconf}->delete_atom($_) foreach (sort {$b <=> $a} @{$tied->{_to_delete}});
    $tied->synchronize();
    $tied->{_frozen} = 0;
}

sub description { 'generic template for key-value association configuration files' }

package Libconf::Glueconf::Generic::KeyValue::Wrapper;
use Libconf::Glueconf::Wrapper;
our @ISA = qw(Libconf::Glueconf::Wrapper);


sub debug { Libconf::debug(@_) }

sub synchronize {
    my $self = shift;
    $self->{hash} = { map { $self->{libconf}->get_atom($_)->{key}, $_ } (0..$self->{libconf}->get_size()-1) };
}

sub TIEHASH {
    my ($pkg, $libconf, $data_description, $data_mapping) = @_;
    debug();
    my $self = bless {
                      libconf => $libconf,
                      _data_description => $data_description,
                      _data_mapping => $data_mapping,
                      _frozen => 0,
                      _to_delete => [],
                      hash => {},
                     }, $pkg;
    $self->synchronize();
    $self
}

sub CLEAR { my ($obj) = @_; debug(); $obj->{hash} = {}; $obj->{libconf}->clear }

sub DELETE {
    my ($obj, $key) = @_; debug("key: $key");
    my $ret;
    defined $obj->{hash}{$key} or return undef;
    if ($obj->{_frozen}) {
        push @{$obj->{_to_delete}}, $obj->{hash}{$key};
    } else {
        $ret = $obj->{libconf}->delete_atom($obj->{hash}{$key});
        delete $obj->{hash}{$key};
    }
    $ret;
}

sub FIRSTKEY { my ($obj) = @_; debug(); keys %{$obj->{hash}}; scalar(each %{$obj->{hash}}) }

sub EXISTS { my ($obj, $key) = @_; debug("key : $key"); exists $obj->{hash}{$key} }

sub NEXTKEY { my ($obj, $lastkey) = @_; debug("lastkey : $lastkey"); scalar(each %{$obj->{hash}}) }

sub STORE {
    my ($obj, $key, $value) = @_; debug("key : $key - value : $value");
    $obj->{_frozen} and warn("I can't store a value in the hash while it is frozen (use thaw to unfreeze)"), return undef;
    $obj->check_type($value, 'SCALAR');
    my $index;
    if (exists $obj->{hash}{$key}) {
        $index = $obj->{hash}{$key};
    } else {
        $index = $obj->{libconf}->append_atom({ type => 'KEY_VALUE', key => $key });
        $obj->{hash}{$key} = $index;
    }
    $obj->store($index, $value);
    $index;
}

sub FETCH {
    my ($obj, $key) = @_;
    debug("key : $key");
    $key eq 'libconf' and return $obj->{libconf};
#    substr($key, 0, 1) eq '_' and return $obj->{$key};
#    $obj->{libconf}->get_atom($obj->{hash}{$key})->{value};
    exists $obj->{hash}{$key} && defined $obj->{hash}{$key} or return undef;
    $obj->fetch($obj->{hash}{$key});
}

1;

