#!/usr/bin/perl

# Author : Damien KROTKINE (dams@gentoo.org)
#
# Contributors :
#
# Copyright (C) 2002 Damien KROTKINE (dams@gentoo.org)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; only version 2, not any other version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

package Libconf::Glueconf::System::Lilo;
use strict;
use vars qw(@ISA);
use Libconf;
use Libconf::Glueconf;
use Libconf::Templates::Lilo;
@ISA = qw(Libconf::Glueconf);

sub new {
    my ($class, $args) = @_;
    my $libconf = new Libconf::Templates::Lilo(init_args($args, { glueconf_template_name => 'System/Lilo' } ));
    $libconf->read_conf();
    $libconf->set_uniq(qw(key sections));

    tie my %wrapper, 'Libconf::Glueconf::System::Lilo::Wrapper', $libconf;
    bless \%wrapper, $class;
}

sub description { 'template for lilo lilo.conf configuration file' }

package Libconf::Glueconf::System::Lilo::Wrapper;
use Libconf qw(:helpers);
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub debug { Libconf::debug(@_) }

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    'global';
}

sub DELETE {
    my ($obj, $key) = @_;
    debug("key: $key");
    member($key, qw(global kernels others)) or die "trying to delete something else than 'global, kernels, others'\n";
    my %hash;
    tie %hash, 'Libconf::Glueconf::System::Lilo::' . ucfirst($key), $obj->{libconf};
    %hash = ();
}

sub EXISTS {
    my ($obj, $key) = @_;
    debug("key : $key");
    scalar member($key, qw(global kernels others));
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    debug("lastkey : $lastkey");
    $lastkey eq 'global' and return 'kernels';
    $lastkey eq 'kernels' and return 'others';
    undef;
}

sub STORE {
    my ($obj, $key, $value) = @_;
    debug("key : $key - value : $value");
    member($key, qw(global kernels others)) or die "trying to store something else than 'global, kernels, others'\n";
    my %hash;
    tie %hash, 'Libconf::Glueconf::System::Lilo::' . ucfirst($key), $obj->{libconf};
    ref $value eq 'HASH' or die 'trying to store something else than hash ref';
    %hash = %$value;
}

sub FETCH {
    my ($obj, $key) = @_;
    debug("key : $key");
    $key eq 'libconf' and return $obj->{libconf};
    member($key, qw(global kernels others)) or die "trying to fetch something else than 'global, kernels, others'\n";
    my %hash;
    tie %hash, 'Libconf::Glueconf::System::Lilo::' . ucfirst($key), $obj->{libconf};
    \%hash;
}

package Libconf::Glueconf::System::Lilo::Global;
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub debug { Libconf::debug(@_) }

sub CLEAR {
    my ($obj) = @_;
    tie my %hash, 'Libconf::Glueconf::System::Lilo::GlobalKeyValueOptions', $obj->{libconf};
    %hash = ();
    tie my @array, 'Libconf::Glueconf::System::Lilo::GlobalValueOptions', $obj->{libconf};
    @array = ();
}

sub DELETE {
    my ($obj, $key) = @_;
    debug("key : $key");
    if ($key eq 'keyvalue') {
        tie my %hash, 'Libconf::Glueconf::System::Lilo::GlobalKeyValueOptions', $obj->{libconf};
        %hash = ();
    }
    if ($key eq 'value') {
        tie my @array, 'Libconf::Glueconf::System::Lilo::GlobalValueOptions', $obj->{libconf};
        @array = ();
    }
    die "key was not one of 'value', 'keyvalue'\n";
}

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    'keyvalue';
}

sub EXISTS {
    my ($obj, $key) = @_;
    $key eq 'value' || $key eq 'keyvalue';
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    $lastkey eq 'keyvalue' and return 'value';
    undef
}

sub STORE {
    my ($obj, $key, $value) = @_;
    if ($key eq 'keyvalue') {
        ref $value eq 'HASH' or die 'trying to store something else than hash ref';
        tie my %hash, 'Libconf::Glueconf::System::Lilo::GlobalKeyValueOptions', $obj->{libconf};
        %hash = %$value;
    }
    if ($key eq 'value') {
        ref $value eq 'ARRAY' or die 'trying to store something else than array ref';
        tie my @array, 'Libconf::Glueconf::System::Lilo::GlobalValueOptions', $obj->{libconf};
        @array = @$value;
    }
    die "wrong key : $key, it should be on of 'value', 'keyvalue'\n";
}

sub FETCH {
    my ($obj, $key) = @_;
    if ($key eq 'keyvalue') {
        tie my %hash, 'Libconf::Glueconf::System::Lilo::GlobalKeyValueOptions', $obj->{libconf};
        return \%hash;
    }
    if ($key eq 'value') {
        tie my @array, 'Libconf::Glueconf::System::Lilo::GlobalValueOptions', $obj->{libconf};
        return \@array;
    }
    die "key was not one of 'value', 'keyvalue'\n";
}

package Libconf::Glueconf::System::Lilo::GlobalKeyValueOptions;
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub _get_atom_list { shift->{libconf}->find_atom_pos( {sections => [ ], type => 'KEY_VALUE', @_ }) }

sub debug { Libconf::debug(@_) }

sub CLEAR {
    my ($obj) = @_;
    debug();
    foreach ($obj->_get_atom_list()) {
        $obj->{libconf}->delete_atom($_);
    }
}

sub DELETE {
    my ($obj, $key) = @_;
    debug("key : $key");
    foreach ($obj->_get_atom_list(key => $key)) {
        $obj->{libconf}->delete_atom($_);
    }
}

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    my @pos = $obj->_get_atom_list();
    @pos or return undef;
    my $atom = $obj->{libconf}->get_atom($pos[0]);
    $atom->{key};
}

sub EXISTS {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->_get_atom_list(key => $key);
    defined $pos ? 1 : 0;
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    debug("Global - NEXTKEY - lastkey : $lastkey");
    my @list = $obj->_get_atom_list();
    my $match;
    foreach (0..$#list) {
        $lastkey eq $obj->{libconf}->get_atom($list[$_])->{key} and $match=$_, last;
    }
    defined $match or return undef;
    defined $list[$match+1] or return undef;
    $obj->{libconf}->get_atom($list[$match+1])->{key};
}

sub STORE {
    my ($obj, $key, $value) = @_;
    debug("key : $key - value : $value");
    $obj->check_type($value, 'SCALAR');
    my $pos = $obj->_get_atom_list(key => $key);
    if (!defined $pos) {
        $pos = $obj->_get_atom_list()+1;
        $obj->{libconf}->insert_atom($pos, { type => 'KEY_VALUE', key => $key, sections => [] });
    }
    $obj->store($pos, $value);
}

sub FETCH {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->_get_atom_list(key => $key);
    defined $pos or return undef;
    $obj->fetch($pos);
}

package Libconf::Glueconf::System::Lilo::GlobalValueOptions;
use Libconf::Glueconf::Wrapper;
our @ISA = qw(Libconf::Glueconf::Wrapper);

sub _get_atom_list { shift->{libconf}->find_atom_pos( {sections => [ ], type => 'VALUE', @_ }) }

sub debug { Libconf::debug(@_) }

sub TIEARRAY {
    my ($pkg, $libconf) = @_;
    debug();
    bless { libconf => $libconf }, $pkg;
}

sub FETCH {
    my ($obj, $idx) = @_;
    debug("idx : $idx");
    my @pos = $obj->_get_atom_list();
    $obj->fetch($pos[$idx]);
}

sub STORE {
    my ($obj, $idx, $value) = @_;
    debug("idx : $idx - value : $value");
    my @pos = $obj->_get_atom_list();
    if ($idx > $#pos) {
        $idx = $pos[-1]+1;
        $obj->{libconf}->insert_atom($idx, {type => 'VALUE'});
    }
    $obj->store($idx, $value);
}

sub FETCHSIZE {
    my ($obj) = @_;
    debug();
    my @pos = $obj->_get_atom_list();
    scalar(@pos);
}

sub STORESIZE {
    my ($obj, $count) = @_;
    debug("count : $count");
    my $size = $obj->FETCHSIZE;
    if ($count > $size) {
        foreach ($count-$size..$count) {
            $obj->STORE($_, '');
        }
    } elsif ($count < $size) {
        foreach (0..$size-$count-2) {
            $obj->POP;
        }
    }
}

sub EXTEND {}

sub EXISTS {
    my ($obj, $index) = @_;
    debug("index : $index");
    defined $obj->FETCH($index);
}

sub CLEAR {
    my $obj = shift;
    debug();
    $obj->{libconf}->delete_atom($_) foreach $obj->_get_atom_list();
}

sub PUSH {
    my $obj = shift;
    my @list = @_;
    debug("list : [" . join('|', @list) . "]");
    my $last = $obj->FETCHSIZE;
    $obj->STORE($last + $_, $list[$_]) foreach (0..$#list);
    $obj->FETCHSIZE;
}

sub POP {
    my $obj = shift;
    debug();
    my @pos = $obj->_get_atom_list();
    my $value = $obj->FETCH($pos[-1]);
    $obj->{libconf}->delete_atom($pos[-1]);
    $value;
}

sub SHIFT {
    my $obj = shift;
    debug();
    my @pos = $obj->_get_atom_list();
    my $value = $obj->FETCH($pos[0]);
    $obj->{libconf}->delete_atom($pos[0]);
    $value;
}

sub UNSHIFT {
    my ($obj, @list) = @_;
    debug("list : [" . join('|', @list));
    my @pos = $obj->_get_atom_list();
    foreach (reverse @list) {
        $obj->{libconf}->insert_atom($pos[0], {type => 'VALUE'});
        $obj->store($pos[0], $_);
    }
}

sub SPLICE {
    my $obj = shift;
    my $offset = shift || 0;
    my $length = shift || $obj->FETCHSIZE - $offset;
    my @list = @_;
    debug("offset : $offset - length : $length - list : [" . join('|', @list) . "]");
    my @pos = $obj->_get_atom_list();
    my @ret;
    foreach (0..$length-1) {
        push @ret, $obj->FETCH($offset+$_);
        exists $list[$_] ? $obj->STORE($offset+$_, $list[$_]) : $obj->{libconf}->delete_atom($pos[$offset+$_]);
    }
    wantarray() ? @ret : $ret[-1];
}


package Libconf::Glueconf::System::Lilo::Kernels;
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub debug { Libconf::debug(@_) }

sub _get_atom_list { shift->{libconf}->find_atom_pos( {sections => [ ], type => 'SECTION', type2 => 'image', @_ }) }

sub DELETE {
    my ($obj, $key) = @_;
    debug("key: $key");
    my @pos = $obj->_get_atom_list(section_name => $key );
    foreach (@pos) {
        $obj->{libconf}->delete_section($_, 1);
    }
}

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    my @pos = $obj->_get_atom_list();
    @pos > 0 or return undef;
    $obj->{libconf}->get_atom($pos[0])->{section_name};
}

sub EXISTS {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->_get_atom_list(section_name => $key );
    defined $pos;
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    debug("lastkey : $lastkey");
    my @pos = $obj->_get_atom_list();
    my $i = 0;
    my $index;
    foreach (@pos) {
        $obj->{libconf}->get_atom($_)->{section_name} eq $lastkey and $index = $i;
        $i++;
    }
    defined $index or return undef;
    $index == $#pos and return undef;
    $obj->{libconf}->get_atom($pos[$index+1])->{section_name};
}

sub STORE {
    my ($obj, $key, $value) = @_;
    debug("key : $key - value : $value");
    ref $value eq 'HASH' or die 'trying to store something else than hash ref';
    my $index;
    my $pos = $obj->_get_atom_list(section_name => $key);
    defined $pos or $pos = $obj->{libconf}->append_atom({ sections => [ ], type => 'SECTION', type2 => 'image', section_name => $key });
    my %hash;
    tie %hash, 'Libconf::Glueconf::System::Lilo::SectionWrapper', $obj->{libconf}, $pos;
    %hash = %$value;
}

sub FETCH {
    my ($obj, $key) = @_;
    debug("key : $key");
    $key eq 'libconf' and return $obj->{libconf};
    substr($key, 0, 1) eq '_' and return $obj->{$key};
    my $section_position = $obj->_get_atom_list(section_name => $key);
    $obj->{libconf}{_fetch_atom_pos} and $obj->{libconf}{_atom_pos} = $section_position;
    defined $section_position or $section_position = $obj->{libconf}->append_atom({ sections => [ ], type => 'SECTION', type2 => 'image', section_name => $key });
    my %ret;
    tie %ret, 'Libconf::Glueconf::System::Lilo::SectionWrapper', $obj->{libconf}, $section_position;
    \%ret;
}

package Libconf::Glueconf::System::Lilo::Others;
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub debug { Libconf::debug(@_) }

sub _get_atom_list { shift->{libconf}->find_atom_pos( {sections => [ ], type => 'SECTION', type2 => 'other', @_ }) }

sub DELETE {
    my ($obj, $key) = @_;
    debug("key: $key");
    my @pos = $obj->_get_atom_list(section_name => $key );
    foreach (@pos) {
        $obj->{libconf}->delete_section($_, 1);
    }
}

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    my @pos = $obj->_get_atom_list();
    @pos > 0 or return undef;
    $obj->{libconf}->get_atom($pos[0])->{section_name};
}

sub EXISTS {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->_get_atom_list(section_name => $key );
    defined $pos;
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    debug("lastkey : $lastkey");
    my @pos = $obj->_get_atom_list();
    my $i = 0;
    my $index;
    foreach (@pos) {
        $obj->{libconf}->get_atom($_)->{section_name} eq $lastkey and $index = $i;
        $i++;
    }
    defined $index or return undef;
    $index == $#pos and return undef;
    $obj->{libconf}->get_atom($pos[$index+1])->{section_name};
}

sub STORE {
    my ($obj, $key, $value) = @_;
    debug("key : $key - value : $value");
    ref $value eq 'HASH' or die 'trying to store something else than hash ref';
    my $index;
    my $pos = $obj->_get_atom_list(section_name => $key);
    defined $pos or $pos = $obj->{libconf}->append_atom({ sections => [ ], type => 'SECTION', type2 => 'image', section_name => $key });
    my %hash;
    tie %hash, 'Libconf::Glueconf::System::Lilo::SectionWrapper', $obj->{libconf}, $pos;
    %hash = %$value;
}

sub FETCH {
    my ($obj, $key) = @_;
    debug("key : $key");
    $key eq 'libconf' and return $obj->{libconf};
    substr($key, 0, 1) eq '_' and return $obj->{$key};
    my $section_position = $obj->_get_atom_list(section_name => $key);
    $obj->{libconf}{_fetch_atom_pos} and $obj->{libconf}{_atom_pos} = $section_position;
    defined $section_position or $section_position = $obj->{libconf}->append_atom({ sections => [ ], type => 'SECTION', type2 => 'image', section_name => $key });
    my %ret;
    tie %ret, 'Libconf::Glueconf::System::Lilo::SectionWrapper', $obj->{libconf}, $section_position;
    \%ret;
}


package Libconf::Glueconf::System::Lilo::SectionWrapper;
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub debug { Libconf::debug(@_) }

sub TIEHASH {
    my ($pkg, $libconf, $pos) = @_;
    debug();
    bless {
           libconf => $libconf,
           sectionpos => $pos,
          }, $pkg;
}

sub CLEAR {
    my ($obj) = @_;
    tie my %hash, 'Libconf::Glueconf::System::Lilo::SectionKeyValueOptions', $obj->{libconf}, $obj->{sectionpos};
    %hash = ();
    tie my @array, 'Libconf::Glueconf::System::Lilo::SectionValueOptions', $obj->{libconf}, $obj->{sectionpos};
    @array = ();
}

sub DELETE {
    my ($obj, $key) = @_;
    debug("key : $key");
    if ($key eq 'keyvalue') {
        tie my %hash, 'Libconf::Glueconf::System::Lilo::SectionKeyValueOptions', $obj->{libconf}, $obj->{sectionpos};
        %hash = ();
    }
    if ($key eq 'value') {
        tie my @array, 'Libconf::Glueconf::System::Lilo::SectionValueOptions', $obj->{libconf}, $obj->{sectionpos};
        @array = ();
    }
    die "key was not one of 'value', 'keyvalue'\n";
}

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    'keyvalue';
}

sub EXISTS {
    my ($obj, $key) = @_;
    $key eq 'value' || $key eq 'keyvalue';
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    $lastkey eq 'keyvalue' and return 'value';
    undef
}

sub STORE {
    my ($obj, $key, $value) = @_;
    if ($key eq 'keyvalue') {
        ref $value eq 'HASH' or die 'trying to store something else than hash ref';
        tie my %hash, 'Libconf::Glueconf::System::Lilo::SectionKeyValueOptions', $obj->{libconf}, $obj->{sectionpos};
        %hash = %$value;
    }
    if ($key eq 'value') {
        ref $value eq 'ARRAY' or die 'trying to store something else than array ref';
        tie my @array, 'Libconf::Glueconf::System::Lilo::SectionValueOptions', $obj->{libconf}, $obj->{sectionpos};
        @array = @$value;
    }
    die "wrong key : $key, it should be on of 'value', 'keyvalue'\n";
}

sub FETCH {
    my ($obj, $key) = @_;
    if ($key eq 'keyvalue') {
        tie my %hash, 'Libconf::Glueconf::System::Lilo::SectionKeyValueOptions', $obj->{libconf}, $obj->{sectionpos};
        return \%hash;
    }
    if ($key eq 'value') {
        tie my @array, 'Libconf::Glueconf::System::Lilo::SectionValueOptions', $obj->{libconf}, $obj->{sectionpos};
        return \@array;
    }
    die "key was not one of 'value', 'keyvalue'\n";
}

package Libconf::Glueconf::System::Lilo::SectionKeyValueOptions;
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub _get_atom_list { my $obj = shift; $obj->{libconf}->find_atom_pos( {sections => [ { name => $obj->{section_name} } ], type => 'KEY_VALUE', @_ }) }

sub debug { Libconf::debug(@_) }

sub TIEHASH {
    my ($pkg, $libconf, $pos) = @_;
    debug();
    bless {
           libconf => $libconf,
           sectionpos => $pos,
           section_name => $libconf->get_atom($pos)->{section_name},
          }, $pkg;
}

sub CLEAR {
    my ($obj) = @_;
    debug();
    foreach ($obj->_get_atom_list()) {
        $obj->{libconf}->delete_atom($_);
    }
}

sub DELETE {
    my ($obj, $key) = @_;
    debug("key : $key");
    foreach ($obj->_get_atom_list(key => $key)) {
        $obj->{libconf}->delete_atom($_);
    }
}

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    my @pos = $obj->_get_atom_list();
    @pos > 0 or return undef;
    my $atom = $obj->{libconf}->get_atom($pos[0]);
    $atom->{key};
}

sub EXISTS {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->_get_atom_list(key => $key);
    defined $pos ? 1 : 0;
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    debug("NEXTKEY - lastkey : $lastkey");
    my @list = $obj->_get_atom_list();
    my $match;
    foreach (0..$#list) {
        $lastkey eq $obj->{libconf}->get_atom($list[$_])->{key} and $match=$_, last;
    }
    defined $match or return undef;
    defined $list[$match+1] or return undef;
    my $ret = $obj->{libconf}->get_atom($list[$match+1])->{key};
    $ret;
}

sub STORE {
    my ($obj, $key, $value) = @_;
    debug("key : $key - value : $value");
    $obj->check_type($value, 'SCALAR');
    my $pos = $obj->_get_atom_list(key => $key);
    if (!defined $pos) {
        $pos = $obj->_get_atom_list()+1;
        $obj->{libconf}->insert_atom($pos, { type => 'KEY_VALUE', key => $key, sections => [ $obj->{section_name} ] });
    }
    $obj->store($pos, $value);
}

sub FETCH {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->_get_atom_list(key => $key);
    defined $pos or return undef;
    $obj->fetch($pos);
}

package Libconf::Glueconf::System::Lilo::SectionValueOptions;
use Libconf::Glueconf::Wrapper;
our @ISA = qw(Libconf::Glueconf::Wrapper);

sub _get_atom_list { my $obj = shift; $obj->{libconf}->find_atom_pos( {sections => [ { name => $obj->{section_name} } ], type => 'VALUE', @_ }) }

sub debug { Libconf::debug(@_) }

sub TIEARRAY {
    my ($pkg, $libconf, $pos) = @_;
    debug();
    bless {
           libconf => $libconf,
           sectionpos => $pos,
           section_name => $libconf->get_atom($pos)->{section_name},
          }, $pkg;
}

sub FETCH {
    my ($obj, $idx) = @_;
    debug("idx : $idx");
    my @pos = $obj->_get_atom_list();
    $obj->fetch($pos[$idx]);
}

sub STORE {
    my ($obj, $idx, $value) = @_;
    debug("idx : $idx - value : $value");
    my @pos = $obj->_get_atom_list();
    if ($idx > $#pos) {
        $idx = $pos[-1]+1;
        $obj->{libconf}->insert_atom($idx, {type => 'VALUE', sections => [ $obj->{section_name} ] });
    }
    $obj->store($idx, $value);
}

sub FETCHSIZE {
    my ($obj) = @_;
    debug();
    my @pos = $obj->_get_atom_list();
    scalar(@pos);
}

sub STORESIZE {
    my ($obj, $count) = @_;
    debug("count : $count");
    my $size = $obj->FETCHSIZE;
    if ($count > $size) {
        foreach ($count-$size..$count) {
            $obj->STORE($_, '');
        }
    } elsif ($count < $size) {
        foreach (0..$size-$count-2) {
            $obj->POP;
        }
    }
}

sub EXTEND {}

sub EXISTS {
    my ($obj, $index) = @_;
    debug("index : $index");
    defined $obj->FETCH($index);
}

sub CLEAR {
    my $obj = shift;
    debug();
    $obj->{libconf}->delete_atom($_) foreach $obj->_get_atom_list();
}

sub PUSH {
    my $obj = shift;
    my @list = @_;
    debug("list : [" . join('|', @list) . "]");
    my $last = $obj->FETCHSIZE;
    $obj->STORE($last + $_, $list[$_]) foreach (0..$#list);
    $obj->FETCHSIZE;
}

sub POP {
    my $obj = shift;
    debug();
    my @pos = $obj->_get_atom_list();
    my $value = $obj->FETCH($pos[-1]);
    $obj->{libconf}->delete_atom($pos[-1]);
    $value;
}

sub SHIFT {
    my $obj = shift;
    debug();
    my @pos = $obj->_get_atom_list();
    my $value = $obj->FETCH($pos[0]);
    $obj->{libconf}->delete_atom($pos[0]);
    $value;
}

sub UNSHIFT {
    my ($obj, @list) = @_;
    debug("list : [" . join('|', @list));
    my @pos = $obj->_get_atom_list();
    foreach (reverse @list) {
        $obj->{libconf}->insert_atom($pos[0], {type => 'VALUE'});
        $obj->store($pos[0], $_);
    }
}

sub SPLICE {
    my $obj = shift;
    my $offset = shift || 0;
    my $length = shift || $obj->FETCHSIZE - $offset;
    my @list = @_;
    debug("offset : $offset - length : $length - list : [" . join('|', @list) . "]");
    my @pos = $obj->_get_atom_list();
    my @ret;
    foreach (0..$length-1) {
        push @ret, $obj->FETCH($offset+$_);
        exists $list[$_] ? $obj->STORE($offset+$_, $list[$_]) : $obj->{libconf}->delete_atom($pos[$offset+$_]);
    }
    wantarray() ? @ret : $ret[-1];
}






























# sub _get_atom_list {
#     $obj->{libconf}->find_atom_pos( {sections => [ ], type => 'KEY_VALUE', @_ }),
#     $obj->{libconf}->find_atom_pos( {sections => [ ], type => 'VALUE', @_ });
# }

# sub CLEAR {
#     my ($obj) = @_;
#     debug();
#     foreach (_get_atom_list()) {
#         $obj->{libconf}->delete_atom($_);
#     }
# }

# sub DELETE {
#     my ($obj, $key) = @_;
#     debug("key : $key");
#     foreach (_get_atom_list(key => $key)) {
#         $obj->{libconf}->delete_atom($_);
#     }
# }

# sub FIRSTKEY {
#     my ($obj) = @_;
#     debug();
#     my @pos = _get_atom_list(key => $key);
#     @pos or return undef;
#     my $atom = $obj->{libconf}->get_atom($pos[0]);
#     $atom->{type} eq 'KEY_VALUE' and return $atom->{key};
#     $atom->{type} eq 'VALUE' and return $atom->{value};
#     undef;
# }

# sub EXISTS {
#     my ($obj, $key) = @_;
#     debug("key : $key");
#     my $pos = _get_atom_list(key => $key);
#     defined $pos ? 1 : 0;
# }

# sub NEXTKEY {
#     my ($obj, $lastkey) = @_;
#     debug("Global - NEXTKEY - lastkey : $lastkey");
#     my @list = _get_atom_list();
#     my $match;
#     foreach (0..$#list) {
#         my $pos = $list[$_];
#         my $atom = $obj->{libconf}->get_atom($pos);
#         my $val2;
#         $atom->{type} eq 'KEY_VALUE' and $val2 = $atom->{key};
#         $atom->{type} eq 'VALUE' and $val2 = $atom->{value};
#         $lastkey eq $val2 and $match=$_, last;
#     }
#     defined $match or return undef;
#     defined $list[$match+1] or return undef;
#     my $atom = $obj->{libconf}->get_atom($list[$match+1]);
#     $atom->{type} eq 'KEY_VALUE' and return $atom->{key};
#     $atom->{type} eq 'VALUE' and return $atom->{value};
#     undef;
# }

# sub STORE {
#     my ($obj, $key, $value) = @_;
#     debug("key : $key - value : $value");
#     $obj->check_type($value, 'SCALAR');
#     my $pos = $obj->{libconf}->find_atom_pos( {type => 'KEY_VALUE', key => $key },
#                                             $obj->{firstatom}+1, $obj->{lastatom}
#                                           );
#     if (!defined $pos) {
#         $obj->{lastatom}++;
#         $obj->{libconf}->insert_atom($obj->{lastatom}, { type => 'KEY_VALUE', key => $key,
#                                                          sections => [ @{$obj->{sections}} ],
#                                                        }
#                                     );
#         $pos = $obj->{lastatom};
#     }
#     $obj->store($pos, $value);
# }

# sub FETCH {
#     my ($obj, $key) = @_;
#     debug("key : $key");
#     my $pos = $obj->{libconf}->find_atom_pos( {type => 'KEY_VALUE', key => $key },
#                                             $obj->{firstatom}+1, $obj->{lastatom}
#                                           );
#     defined $pos or return undef;
# #    $obj->fetch($obj->{libconf}->get_atom($pos), 'value');
#     $obj->fetch($pos);
# }

1;

